import jax
import jax.numpy as jnp
from flax import linen as nn
import chex
from typing import Tuple, Optional
from .shared import (
    identity_out,
    tanh_out,
    categorical_out,
    gaussian_out,
    default_bias_init,
    kernel_init_fn,
)


def conv_relu_block(
    x: chex.Array,
    features: int,
    kernel_size: Tuple[int, int],
    strides: Tuple[int, int],
    padding: str = "SAME",
    kernel_init_type: str = "lecun_normal",
) -> chex.Array:
    """Convolution layer + ReLU activation."""
    x = nn.Conv(
        features=features,
        kernel_size=kernel_size,
        strides=strides,
        use_bias=True,
        padding=padding,
        bias_init=default_bias_init(),
        kernel_init=kernel_init_fn[kernel_init_type](),
    )(x)
    x = nn.relu(x)
    return x


def conv_relu_pool_block(
    x: chex.Array,
    features: int,
    kernel_size: Tuple[int, int],
    strides: Tuple[int, int],
    padding: str = "SAME",
    kernel_init_type: str = "lecun_normal",
) -> chex.Array:
    """Convolution layer + ReLU activation + Avg. Pooling."""
    x = nn.Conv(
        features=features,
        kernel_size=kernel_size,
        strides=strides,
        use_bias=True,
        padding=padding,
        bias_init=default_bias_init(),
        kernel_init=kernel_init_fn[kernel_init_type](),
    )(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    return x


class CNN(nn.Module):
    """Simple CNN Wrapper with Conv-ReLu-AvgPool blocks."""

    num_output_units: int = 10
    depth_1: int = 1
    depth_2: int = 1
    features_1: int = 8
    features_2: int = 16
    kernel_1: int = 5
    kernel_2: int = 5
    strides_1: int = 1
    strides_2: int = 1
    num_linear_layers: int = 1
    num_hidden_units: int = 16
    hidden_activation: str = "relu"
    output_activation: str = "identity"
    kernel_init_type: str = "lecun_normal"
    model_name: str = "CNN"

    @nn.compact
    def __call__(
        self, x: chex.Array, rng: Optional[chex.PRNGKey] = None
    ) -> chex.Array:
        # Add batch dimension if only processing single 3d array
        if len(x.shape) < 4:
            x = jnp.expand_dims(x, 0)
            batch_case = False
        else:
            batch_case = True

        # Block In 1:
        for _ in range(self.depth_1):
            x = conv_relu_pool_block(
                x,
                self.features_1,
                (self.kernel_1, self.kernel_1),
                (self.strides_1, self.strides_1),
                kernel_init_type=self.kernel_init_type,
            )

        # Block In 2:
        for _ in range(self.depth_2):
            x = conv_relu_pool_block(
                x,
                self.features_2,
                (self.kernel_2, self.kernel_2),
                (self.strides_2, self.strides_2),
                kernel_init_type=self.kernel_init_type,
            )
        # Flatten the output into vector for Dense Readout
        x = x.reshape(x.shape[0], -1)
        # Squeeze and linear layers
        for _ in range(self.num_linear_layers):
            x = nn.Dense(
                features=self.num_hidden_units,
                bias_init=default_bias_init(),
                kernel_init=kernel_init_fn[self.kernel_init_type](),
            )(x)
            if self.hidden_activation == "relu":
                x = nn.relu(x)
            elif self.hidden_activation == "tanh":
                x = nn.tanh(x)
            elif self.hidden_activation == "gelu":
                x = nn.gelu(x)
            elif self.hidden_activation == "softplus":
                x = nn.softplus(x)

        if self.output_activation == "identity":
            x = identity_out(x, self.num_output_units, self.kernel_init_type)
        elif self.output_activation == "tanh":
            x = tanh_out(x, self.num_output_units, self.kernel_init_type)
        # Categorical and gaussian output heads require rng for sampling
        elif self.output_activation == "categorical":
            x = categorical_out(
                rng, x, self.num_output_units, self.kernel_init_type
            )
        elif self.output_activation == "gaussian":
            x = gaussian_out(
                rng, x, self.num_output_units, self.kernel_init_type
            )
        # Squeeze away extra dimension - e.g. single action output for RL
        if not batch_case:
            return x.squeeze()
        else:
            return x


class All_CNN_C(nn.Module):
    """All-CNN-inspired architecture as in Springenberg et al. (2015).
    Reference: https://arxiv.org/abs/1412.6806"""

    num_output_units: int = 10
    depth_1: int = 1
    depth_2: int = 1
    features_1: int = 8
    features_2: int = 16
    kernel_1: int = 5
    kernel_2: int = 5
    strides_1: int = 1
    strides_2: int = 1
    final_window: Tuple[int, int] = (28, 28)
    output_activation: str = "identity"
    kernel_init_type: str = "lecun_normal"
    model_name: str = "All_CNN_C"

    @nn.compact
    def __call__(
        self, x: chex.Array, rng: Optional[chex.PRNGKey] = None
    ) -> chex.Array:
        # Add batch dimension if only processing single 3d array
        if len(x.shape) < 4:
            x = jnp.expand_dims(x, 0)
            batch_case = False
        else:
            batch_case = True
        # Block In 1:
        for _ in range(self.depth_1):
            x = conv_relu_block(
                x,
                self.features_1,
                (self.kernel_1, self.kernel_1),
                (self.strides_1, self.strides_1),
                kernel_init_type=self.kernel_init_type,
            )

        # Block In 2:
        for _ in range(self.depth_2):
            x = conv_relu_block(
                x,
                self.features_2,
                (self.kernel_2, self.kernel_2),
                (self.strides_2, self.strides_2),
                kernel_init_type=self.kernel_init_type,
            )

        # Block Out: 1 × 1 conv. num_outputs-ReLu ×n
        x = nn.Conv(
            features=self.num_output_units,
            kernel_size=(1, 1),
            strides=(1, 1),
            use_bias=True,
            padding="SAME",
            bias_init=default_bias_init(),
            kernel_init=kernel_init_fn[self.kernel_init_type](),
        )(x)

        # Global average pooling -> logits
        x = nn.avg_pool(
            x, window_shape=self.final_window, strides=None, padding="VALID"
        )
        x = jnp.squeeze(x, axis=(1, 2))

        if self.output_activation == "tanh":
            x = nn.tanh(x)
        # Categorical head requires rng for sampling
        elif self.output_activation == "categorical":
            x = jax.random.categorical(rng, x)
        # No gaussian option implemented so far - need second 1x1 conv + pool
        # Squeeze away extra dimension - e.g. single action output for RL
        if not batch_case:
            return x.squeeze()
        else:
            return x



from functools import partial
import jax.numpy as jnp
from flax import linen as nn
import chex
from .evo_modules import (
    CompressionPerceiver,
    SolutionPerceiver,
    DistributionAttention,
    DistributionUpdateNetwork,
)
from .attention import AttentionEncoder


class EvoTransformer(nn.Module):
    embed_dim: int
    num_heads: int
    num_latents: int
    latent_dim: int
    num_layers: int = 1
    dropout_prob: float = 0.0
    input_dropout_prob: float = 0.0
    use_bias: bool = False
    use_fitness_encoder: bool = True
    use_dist_encoder: bool = True
    use_crossd_encoder: bool = True
    out_att_maps: bool = False

    def setup(self):
        self.sol_perceiver = SolutionPerceiver(
            num_latents=self.num_latents,
            latent_dim=self.latent_dim,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_layers=self.num_layers,
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )
        self.fit_perceiver = CompressionPerceiver(
            num_latents=self.num_latents,
            latent_dim=self.latent_dim,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_layers=self.num_layers,
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )
        self.dist_encoder = DistributionAttention(
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_layers=self.num_layers,
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )
        self.crossd_encoder = CompressionPerceiver(
            num_latents=self.num_latents,
            latent_dim=self.latent_dim,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_layers=1,  # For now only with a single layer possible
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )
        self.transformer_backbone = partial(
            AttentionEncoder,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_layers=self.num_layers,
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )
        self.dist_update = partial(
            DistributionUpdateNetwork,
            embed_dim=self.embed_dim,
            dropout_prob=self.dropout_prob,
            use_bias=self.use_bias,
        )
        self.lift_transformer = nn.vmap(
            self.transformer_backbone,
            variable_axes={"params": None},
            split_rngs={"params": False, "dropout": False},
            in_axes=(0, None, None, None),
            out_axes=0,
        )
        self.lift_distribution = nn.vmap(
            self.dist_update,
            variable_axes={"params": None},
            split_rngs={"params": False, "dropout": False},
            in_axes=(0, None),
            out_axes=0,
        )

    @nn.compact
    def __call__(
        self,
        solution_features: chex.Array,
        fitness_features: chex.Array,
        dist_features: chex.Array,
        mask=None,
        add_positional_encoding=True,
        train=False,
        verbose=False,
    ):
        batch_size, seq_len, popsize, num_dims, feature_dim = solution_features.shape
        sol_encoding, sol_att = self.sol_perceiver(solution_features, train)
        if self.use_fitness_encoder:
            fit_encoding, fit_att = self.fit_perceiver(fitness_features, train)
        if self.use_dist_encoder:
            dist_encoding, dist_att = self.dist_encoder(dist_features, train)
        if self.use_crossd_encoder:
            crossd_encoding, crossd_att = self.crossd_encoder(dist_features, train)

        if verbose:
            print("Solution encoding shape:", sol_encoding.shape)
            if self.use_fitness_encoder:
                print("Fitness encoding shape:", fit_encoding.shape)
            if self.use_dist_encoder:
                print("Dist encoding shape:", dist_encoding.shape)
            if self.use_crossd_encoder:
                print("Crossd encoding shape:", crossd_encoding.shape)

        sol_encoding = sol_encoding.transpose(3, 0, 1, 2, 4)
        if verbose:
            print("Solution encoding shape after reshape:", sol_encoding.shape)

        if self.use_fitness_encoder:
            fit_encoding = jnp.repeat(
                fit_encoding.reshape(1, *fit_encoding.shape), num_dims, axis=0
            )
            if verbose:
                print("Fitness encoding shape after reshape:", fit_encoding.shape)

        if self.use_dist_encoder:
            dist_encoding = dist_encoding.transpose(2, 0, 1, 3)
            if verbose:
                print("Dist encoding shape after reshape:", dist_encoding.shape)

        if self.use_crossd_encoder:
            crossd_encoding = jnp.repeat(
                crossd_encoding.reshape(1, *crossd_encoding.shape), num_dims, axis=0
            )
            if verbose:
                print("Crossd encoding shape after reshape:", crossd_encoding.shape)

        combined_encoding = sol_encoding
        if self.use_fitness_encoder:
            combined_encoding = jnp.concatenate(
                [combined_encoding, fit_encoding], axis=-1
            )
        combined_encoding = combined_encoding.reshape(num_dims, batch_size, seq_len, -1)
        if verbose:
            print("Combined encoding shape:", combined_encoding.shape)

        if self.use_dist_encoder:
            combined_encoding = jnp.concatenate(
                [combined_encoding, dist_encoding], axis=-1
            )
            if verbose:
                print("Combined encoding shape after dist:", combined_encoding.shape)

        if self.use_crossd_encoder:
            crossd_encoding = crossd_encoding.reshape(num_dims, batch_size, seq_len, -1)
            combined_encoding = jnp.concatenate(
                [combined_encoding, crossd_encoding], axis=-1
            )
            if verbose:
                print("Combined encoding shape after crossd:", combined_encoding.shape)

        out, mhsa_att = self.lift_transformer(name="DimBatchedTransformer")(
            combined_encoding, mask, add_positional_encoding, train
        )
        if self.out_att_maps:
            mhsa_att = jnp.array(mhsa_att).transpose(2, 1, 0, 3, 4, 5)
        if verbose:
            print("Transformer output shape:", out.shape)

        distrib_out = self.lift_distribution(name="DistributionUpdate")(out, train)
        distrib_out = distrib_out.transpose(3, 1, 2, 0)
        if verbose:
            print("Distribution output shape:", distrib_out.shape)

        # Collect all attention maps
        all_att_out = {"solution": sol_att, "time": mhsa_att}
        if self.use_fitness_encoder:
            all_att_out["fitness"] = fit_att
        if self.use_dist_encoder:
            all_att_out["distribution"] = dist_att
        if self.use_crossd_encoder:
            all_att_out["cross_dim"] = crossd_att
        return distrib_out, all_att_out



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    archive: chex.Array
    fitness: chex.Array
    copy_id: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    noise_scale: float = 0.1
    truncation_selection: float = 0.2
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class PBT(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Synchronous Population-Based Training (Jaderberg et al., 2017)
        Reference: https://arxiv.org/abs/1711.09846"""
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.strategy_name = "PBT"

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams()

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """
        `initialize` the differential evolution strategy.
        """
        initialization = jax.random.uniform(
            rng,
            (self.popsize, self.num_dims),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            archive=initialization,
            fitness=jnp.zeros(self.popsize) - 20e10,
            copy_id=jnp.zeros(self.popsize, dtype=jnp.int32),
            best_member=jnp.zeros(self.num_dims),
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """
        `ask` for new proposed candidates to evaluate next.
        Perform explore-exploit step.
        1) Check exploit criterion (e.g. in top 20% of performer).
        2) If not exploit: Copy hyperparams from id and explore/perturb around.
        3) Return new hyperparameters and copy_id (same if exploit)
        """
        rng_members = jax.random.split(rng, self.popsize)
        member_ids = jnp.arange(self.popsize)
        exploit_bool, copy_id, hyperparams = jax.vmap(
            single_member_exploit, in_axes=(0, None, None, None)
        )(member_ids, state.archive, state.fitness, params)
        hyperparams = jax.vmap(single_member_explore, in_axes=(0, 0, 0, None))(
            rng_members, exploit_bool, hyperparams, params
        )
        return hyperparams, state.replace(copy_id=copy_id)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state. - Only copy if improved performance."""
        replace = fitness >= state.fitness
        archive = (
            jnp.expand_dims(replace, 1) * x
            + (1 - jnp.expand_dims(replace, 1)) * state.archive
        )
        fitness = replace * fitness + (1 - replace) * state.fitness
        return state.replace(archive=archive, fitness=fitness)


def single_member_exploit(
    member_id: int,
    archive: chex.Array,
    fitness: chex.Array,
    params: EvoParams,
) -> Tuple[bool, int, chex.Array]:
    """Get the top and bottom performers."""
    best_id = jnp.argmax(fitness)
    exploit_bool = member_id != best_id  # Copy if worker not best
    copy_id = jax.lax.select(exploit_bool, best_id, member_id)
    hyperparams_copy = archive[copy_id]
    return exploit_bool, copy_id, hyperparams_copy


def single_member_explore(
    rng: chex.PRNGKey,
    exploit_bool: int,
    hyperparams: chex.Array,
    params: EvoParams,
) -> chex.Array:
    """Perform multiplicative noise exploration."""
    explore_noise = (
        jax.random.normal(rng, hyperparams.shape) * params.noise_scale
    )
    hyperparams_explore = jax.lax.select(
        exploit_bool, hyperparams + explore_noise, hyperparams
    )
    return hyperparams_explore



import functools
from typing import Tuple
import chex
from flax import struct
import jax
import jax.numpy as jnp
from evosax.strategies.des import get_des_weights
from evosax.strategies.snes import get_snes_weights
from evosax.core.fitness import (
    centered_rank_trafo,
    z_score_trafo,
    compute_l2_norm,
    range_norm_trafo,
)


@struct.dataclass
class FitnessFeaturesState:
    best_fitness: float


class FitnessFeaturizer(object):
    def __init__(
        self,
        popsize: int,
        num_dims: int,
        seq_len: int,
        improved_best: bool = False,
        z_score: bool = False,
        norm_diff_best: bool = False,
        norm_range: bool = False,
        snes_weights: bool = False,
        des_weights: bool = False,
        w_decay: float = 0.0,
        maximize: bool = False,
        verbose: bool = False,
    ):
        self.popsize = popsize
        self.num_dims = num_dims
        self.seq_len = seq_len
        self.improved_best = improved_best
        self.z_score = z_score
        self.norm_diff_best = norm_diff_best
        self.norm_range = norm_range
        self.snes_weights = snes_weights
        self.des_weights = des_weights
        self.w_decay = w_decay
        self.maximize = maximize
        self.verbose = verbose

        if self.verbose:
            print(
                f"Fitness Features / Batch shape: {self.num_features} / {self.example_batch_shape}"
            )
            print("[BASE] Centered ranks in [-0.5, 0.5]")
            print(f"[{self.improved_best}] Improved best fitness -> f < f_best")
            print(f"[{self.z_score}] Z-score normalization -> (f-f_mu)/f_sigma")
            print(
                f"[{self.norm_diff_best}] Normalized difference to best -> (f-f_best) in [-1, 1]"
            )
            print(f"[{self.norm_range}] Normalized fitness -> f in [-1, 1]")
            print(f"[{self.snes_weights}] SNES weights")
            print(f"[{self.des_weights}] DES weights")
            print(f"[{self.w_decay}] Scaled Weight Norm")

    @functools.partial(jax.jit, static_argnums=0)
    def featurize(
        self, x: chex.Array, fitness: chex.Array, state: FitnessFeaturesState
    ) -> Tuple[chex.Array, FitnessFeaturesState]:
        fitness = jax.lax.select(self.maximize, -1 * fitness, fitness)
        fit_out = centered_rank_trafo(fitness).reshape(-1, 1)

        if self.improved_best:
            fit_improve = ((fitness < state.best_fitness) * 1.0).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_improve], axis=1)

        if self.z_score:
            fit_zs = z_score_trafo(fitness).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_zs], axis=1)

        if self.norm_diff_best:
            fit_best = get_norm_diff_best(fitness, state.best_fitness).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_best], axis=1)

        if self.norm_range:
            fit_norm = range_norm_trafo(fitness, -0.5, 0.5).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_norm], axis=1)

        if self.snes_weights:
            fit_snes = get_snes_weights(fitness.shape[0])[fitness.argsort()].reshape(
                -1, 1
            )
            fit_out = jnp.concatenate([fit_out, fit_snes], axis=1)

        if self.des_weights:
            fit_des = get_des_weights(fitness.shape[0])[fitness.argsort()].reshape(
                -1, 1
            )
            fit_out = jnp.concatenate([fit_out, fit_des], axis=1)

        if self.w_decay:
            fit_wnorm = compute_l2_norm(x).reshape(-1, 1)
            fit_out = jnp.concatenate([fit_out, fit_wnorm], axis=1)

        best_fitness = update_best_fitness(fitness, state.best_fitness, self.maximize)
        return fit_out, FitnessFeaturesState(best_fitness=best_fitness)

    @functools.partial(jax.jit, static_argnums=0)
    def initialize(self) -> FitnessFeaturesState:
        return FitnessFeaturesState(best_fitness=jnp.finfo(jnp.float32).max)

    @property
    def num_features(self) -> int:
        return (
            1
            + self.improved_best
            + self.z_score
            + self.norm_diff_best
            + self.norm_range
            + self.snes_weights
            + self.des_weights
            + (self.w_decay > 0.0)
        )

    @property
    def example_batch_shape(self) -> Tuple[int, ...]:
        return (
            1,  # batchsize
            self.seq_len,  # timesteps
            self.popsize,  # popsize
            self.num_features,
        )


def update_best_fitness(
    fitness: chex.Array, best_fitness: float, maximize: bool = False
) -> chex.Array:
    fitness_min = jax.lax.select(maximize, -1 * fitness, fitness)
    best_fit_min = jax.lax.select(maximize, -1 * best_fitness, best_fitness)
    best_in_gen = jnp.argmin(fitness_min)
    best_in_gen_fit = fitness_min[best_in_gen]
    replace_fit = best_in_gen_fit < best_fit_min
    best_fitness = jax.lax.select(replace_fit, best_in_gen_fit, best_fit_min)
    best_fitness = jax.lax.select(maximize, -1 * best_fitness, best_fitness)
    return best_fitness


def get_norm_diff_best(fitness: chex.Array, best_fitness: float) -> chex.Array:
    fitness = jnp.clip(fitness, -1e10, 1e10)
    diff_best = fitness - best_fitness
    return jnp.clip(
        diff_best / (jnp.nanmax(diff_best) - jnp.nanmin(diff_best) + 1e-10),
        -1,
        1,
    )



# Import additional utilities for reshaping flat parameters into net dict
from .reshape import ParameterReshaper, ravel_pytree

# Import additional utilities for reshaping fitness
from .fitness import FitnessShaper

# Import Gradient Based Optimizer step functions
from .optimizer import (
    SGD,
    Adam,
    RMSProp,
    ClipUp,
    Adan,
    OptState,
    OptParams,
    exp_decay,
)

GradientOptimizer = {
    "sgd": SGD,
    "adam": Adam,
    "rmsprop": RMSProp,
    "clipup": ClipUp,
    "adan": Adan,
}


__all__ = [
    "ParameterReshaper",
    "ravel_pytree",
    "FitnessShaper",
    "GradientOptimizer",
    "SGD",
    "Adam",
    "RMSProp",
    "ClipUp",
    "Adan",
    "OptState",
    "OptParams",
    "exp_decay",
]



from .simple_ga import SimpleGA
from .simple_es import SimpleES
from .cma_es import CMA_ES
from .de import DE
from .pso import PSO
from .open_es import OpenES
from .pgpe import PGPE
from .pbt import PBT
from .persistent_es import PersistentES
from .ars import ARS
from .sep_cma_es import Sep_CMA_ES
from .bipop_cma_es import BIPOP_CMA_ES
from .ipop_cma_es import IPOP_CMA_ES
from .full_iamalgam import Full_iAMaLGaM
from .indep_iamalgam import Indep_iAMaLGaM
from .ma_es import MA_ES
from .lm_ma_es import LM_MA_ES
from .rm_es import RmES
from .gld import GLD
from .sim_anneal import SimAnneal
from .snes import SNES
from .xnes import xNES
from .esmc import ESMC
from .des import DES
from .samr_ga import SAMR_GA
from .gesmr_ga import GESMR_GA
from .guided_es import GuidedES
from .asebo import ASEBO
from .cr_fm_nes import CR_FM_NES
from .mr15_ga import MR15_GA
from .random import RandomSearch
from .les import LES
from .lga import LGA
from .noise_reuse_es import NoiseReuseES
from .hill_climber import HillClimber
from .evotf_es import EvoTF_ES

__all__ = [
    "SimpleGA",
    "SimpleES",
    "CMA_ES",
    "DE",
    "PSO",
    "OpenES",
    "PGPE",
    "PBT",
    "PersistentES",
    "ARS",
    "Sep_CMA_ES",
    "BIPOP_CMA_ES",
    "IPOP_CMA_ES",
    "Full_iAMaLGaM",
    "Indep_iAMaLGaM",
    "MA_ES",
    "LM_MA_ES",
    "RmES",
    "GLD",
    "SimAnneal",
    "SNES",
    "xNES",
    "ESMC",
    "DES",
    "SAMR_GA",
    "GESMR_GA",
    "GuidedES",
    "ASEBO",
    "CR_FM_NES",
    "MR15_GA",
    "RandomSearch",
    "LES",
    "LGA",
    "NoiseReuseES",
    "HillClimber",
    "EvoTF_ES",
]



import jax
import jax.numpy as jnp
from evosax import CMA_ES, ARS, ParameterReshaper, NetworkMapper
from evosax.problems import (
    BBOBFitness,
    GymnaxFitness,
    VisionFitness,
    SequenceFitness,
)


def test_classic_rollout(classic_name: str):
    rng = jax.random.PRNGKey(0)
    evaluator = BBOBFitness(classic_name, num_dims=2)
    strategy = CMA_ES(popsize=20, num_dims=2, elite_ratio=0.5)
    params = strategy.default_params
    state = strategy.initialize(rng, params)

    # Run the ask-eval-tell loop
    rng, rng_gen, rng_eval = jax.random.split(rng, 3)
    x, state = strategy.ask(rng_gen, state, params)
    fitness = evaluator.rollout(rng_eval, x)
    assert fitness.shape == (20,)


def test_env_ffw_rollout(env_name: str):
    rng = jax.random.PRNGKey(0)
    evaluator = GymnaxFitness(env_name, num_env_steps=100, num_rollouts=10)
    network = NetworkMapper["MLP"](
        num_hidden_units=64,
        num_hidden_layers=2,
        num_output_units=evaluator.action_shape,
        hidden_activation="relu",
        output_activation="categorical",
    )
    pholder = jnp.zeros((1, evaluator.input_shape[0]))
    net_params = network.init(
        rng,
        x=pholder,
        rng=rng,
    )
    reshaper = ParameterReshaper(net_params)
    evaluator.set_apply_fn(network.apply)

    strategy = ARS(popsize=20, num_dims=reshaper.total_params, elite_ratio=0.5)
    state = strategy.initialize(rng)
    # Run the ask-eval-tell loop
    rng, rng_gen, rng_eval = jax.random.split(rng, 3)
    x, state = strategy.ask(rng_gen, state)
    x_re = reshaper.reshape(x)
    fitness = evaluator.rollout(rng_eval, x_re)

    # Assert shape (#popmembers, #rollouts)
    assert fitness.shape == (20, 10)


# def test_env_rec_rollout(env_name: str):
#     rng = jax.random.PRNGKey(0)
#     evaluator = GymnaxFitness(env_name, num_env_steps=100, num_rollouts=10)
#     network = NetworkMapper["LSTM"](
#         num_hidden_units=64,
#         num_output_units=evaluator.action_shape,
#         output_activation="categorical",
#     )

#     pholder = jnp.zeros((1, evaluator.input_shape[0]))
#     carry_init = network.initialize_carry()
#     net_params = network.init(
#         rng,
#         x=pholder,
#         carry=carry_init,
#         rng=rng,
#     )
#     reshaper = ParameterReshaper(net_params)
#     evaluator.set_apply_fn(network.apply, network.initialize_carry)
#     strategy = ARS(popsize=20, num_dims=reshaper.total_params, elite_ratio=0.5)
#     state = strategy.initialize(rng)

#     # Run the ask-eval-tell loop
#     rng, rng_gen, rng_eval = jax.random.split(rng, 3)
#     x, state = strategy.ask(rng_gen, state)
#     x_re = reshaper.reshape(x)
#     fitness = evaluator.rollout(rng_eval, x_re)

#     # Assert shape (#popmembers, #rollouts)
#     assert fitness.shape == (20, 10)


def test_vision_fitness():
    rng = jax.random.PRNGKey(0)
    evaluator = VisionFitness("MNIST", 4, test=True)
    network = NetworkMapper["CNN"](
        depth_1=1,
        depth_2=1,
        features_1=8,
        features_2=16,
        kernel_1=5,
        kernel_2=5,
        strides_1=1,
        strides_2=1,
        num_linear_layers=0,
        num_output_units=10,
    )
    # Channel last configuration for conv!
    pholder = jnp.zeros((1, 28, 28, 1))
    net_params = network.init(
        rng,
        x=pholder,
        rng=rng,
    )

    reshaper = ParameterReshaper(net_params)
    evaluator.set_apply_fn(network.apply)

    strategy = ARS(popsize=4, num_dims=reshaper.total_params, elite_ratio=0.5)
    state = strategy.initialize(rng)

    # Run the ask-eval-tell loop
    rng, rng_gen, rng_eval = jax.random.split(rng, 3)
    x, state = strategy.ask(rng_gen, state)
    x_re = reshaper.reshape(x)
    loss, acc = evaluator.rollout(rng_eval, x_re)
    assert loss.shape == (4, 1)
    assert acc.shape == (4, 1)


# def test_sequence_fitness():
#     rng = jax.random.PRNGKey(0)
#     evaluator = SequenceFitness(task_name="Addition", batch_size=10, test=False)
#     network = NetworkMapper["LSTM"](
#         num_hidden_units=100,
#         num_output_units=evaluator.action_shape,
#     )
#     params = network.init(
#         rng,
#         x=jnp.ones([1, evaluator.input_shape[0]]),
#         carry=network.initialize_carry(),
#         rng=rng,
#     )
#     param_reshaper = ParameterReshaper(params)
#     evaluator.set_apply_fn(network.apply, network.initialize_carry)

#     strategy = ARS(4, param_reshaper.total_params)
#     es_state = strategy.initialize(rng)

#     x, es_state = strategy.ask(rng, es_state)
#     reshaped_params = param_reshaper.reshape(x)
#     # Rollout population performance, reshape fitness & update strategy.
#     loss, perf = evaluator.rollout(rng, reshaped_params)
#     assert loss.shape == (4, 1)
#     assert perf.shape == (4, 1)



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Optional, Union
from functools import partial
from flax import struct
from .utils import get_best_fitness_member
from .core import ParameterReshaper, FitnessShaper


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    best_member: chex.Array
    best_fitness: float
    gen_counter: int


@struct.dataclass
class EvoParams:
    sigma_init: float = 0.03
    sigma_decay: float = 0.999
    sigma_limit: float = 0.01
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class Strategy(object):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Base Class for an Evolution Strategy."""
        self.popsize = popsize

        # Setup optional parameter reshaper
        self.use_param_reshaper = pholder_params is not None
        if self.use_param_reshaper:
            self.param_reshaper = ParameterReshaper(pholder_params, n_devices)
            self.num_dims = self.param_reshaper.total_params
        else:
            self.num_dims = num_dims
        assert (
            self.num_dims is not None
        ), "Provide either num_dims or pholder_params to strategy."

        # Mean exponential decay coefficient m' = coeff * m
        # Implements form of weight decay regularization
        self.mean_decay = mean_decay
        self.use_mean_decay = mean_decay > 0.0

        # Setup optional fitness shaper
        self.fitness_shaper = FitnessShaper(**fitness_kwargs)

    @property
    def default_params(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        params = self.params_strategy
        return params

    @partial(jax.jit, static_argnums=(0,))
    def initialize(
        self,
        rng: chex.PRNGKey,
        params: Optional[EvoParams] = None,
        init_mean: Optional[Union[chex.Array, chex.ArrayTree]] = None,
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        # Initialize strategy based on strategy-specific initialize method
        state = self.initialize_strategy(rng, params)

        if init_mean is not None:
            state = self.set_mean(state, init_mean)
        return state

    @partial(jax.jit, static_argnums=(0,))
    def ask(
        self,
        rng: chex.PRNGKey,
        state: EvoState,
        params: Optional[EvoParams] = None,
    ) -> Tuple[Union[chex.Array, chex.ArrayTree], EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        # Generate proposal based on strategy-specific ask method
        x, state = self.ask_strategy(rng, state, params)
        # Clip proposal candidates into allowed range
        x_clipped = jnp.clip(x, params.clip_min, params.clip_max)

        # Reshape parameters into pytrees
        if self.use_param_reshaper:
            x_out = self.param_reshaper.reshape(x_clipped)
        else:
            x_out = x_clipped
        return x_out, state

    @partial(jax.jit, static_argnums=(0,))
    def tell(
        self,
        x: Union[chex.Array, chex.ArrayTree],
        fitness: chex.Array,
        state: EvoState,
        params: Optional[EvoParams] = None,
    ) -> chex.ArrayTree:
        """`tell` performance data for strategy state update."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        # Flatten params if using param reshaper for ES update
        if self.use_param_reshaper:
            x = self.param_reshaper.flatten(x)

        # Perform fitness reshaping inside of strategy tell call (if desired)
        fitness_re = self.fitness_shaper.apply(x, fitness)

        # Update the search state based on strategy-specific update
        state = self.tell_strategy(x, fitness_re, state, params)

        # Check if there is a new best member & update trackers
        best_member, best_fitness = get_best_fitness_member(
            x, fitness, state, self.fitness_shaper.maximize
        )

        # Exponentially decay mean if coefficient > 0.0
        if self.use_mean_decay:
            state = state.replace(mean=(1 - self.mean_decay) * state.mean)

        return state.replace(
            best_member=best_member,
            best_fitness=best_fitness,
            gen_counter=state.gen_counter + 1,
        )

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """Search-specific `initialize` method. Returns initial state."""
        raise NotImplementedError

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """Search-specific `ask` request. Returns proposals & updated state."""
        raise NotImplementedError

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """Search-specific `tell` update. Returns updated state."""
        raise NotImplementedError

    def get_eval_params(self, state: EvoState):
        """Return reshaped parameters to evaluate."""
        if self.use_param_reshaper:
            x_out = self.param_reshaper.reshape_single(state.mean)
        else:
            x_out = state.mean
        return x_out

    def set_mean(
        self, state: EvoState, replace_mean: Union[chex.Array, chex.ArrayTree]
    ) -> EvoState:
        if self.use_param_reshaper:
            replace_mean = self.param_reshaper.flatten_single(replace_mean)
        else:
            replace_mean = jnp.asarray(replace_mean)
        state = state.replace(mean=replace_mean)
        return state



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Optional
from functools import partial
from .termination import min_gen_criterion
from flax import struct


@struct.dataclass
class WrapperState:
    strategy_state: chex.ArrayTree
    restart_state: chex.ArrayTree


@struct.dataclass
class WrapperParams:
    strategy_params: chex.ArrayTree
    restart_params: chex.ArrayTree


@struct.dataclass
class RestartState:
    restart_counter: int
    restart_next: bool


@struct.dataclass
class RestartParams:
    min_num_gens: int = 50


class RestartWrapper(object):
    def __init__(self, base_strategy, stop_criteria=[]):
        """Base Class for a Restart Strategy."""
        self.base_strategy = base_strategy
        self.stop_criteria = stop_criteria

    @property
    def num_dims(self) -> int:
        """Get number of problem dimensions from base strategy."""
        return self.base_strategy.num_dims

    @property
    def popsize(self) -> int:
        """Get population size from base strategy."""
        return self.base_strategy.popsize

    @property
    def default_params(self) -> WrapperParams:
        """Return default parameters of evolution strategy."""
        return WrapperParams(
            strategy_params=self.base_strategy.default_params,
            restart_params=self.restart_params,
        )

    @property
    def restart_params(self) -> RestartParams:
        """Return default parameters for strategy restarting."""
        return RestartParams()

    @partial(jax.jit, static_argnums=(0,))
    def initialize(
        self, rng: chex.PRNGKey, params: Optional[WrapperParams] = None
    ) -> WrapperState:
        """`initialize` the evolution strategy."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        strategy_state = self.base_strategy.initialize(
            rng, params.strategy_params
        )
        restart_state = RestartState(restart_counter=0, restart_next=False)
        return WrapperState(strategy_state, restart_state)

    # @partial(jax.jit, static_argnums=(0,))
    def ask(
        self,
        rng: chex.PRNGKey,
        state: WrapperState,
        params: Optional[WrapperParams] = None,
    ) -> Tuple[chex.Array, WrapperState]:
        """`ask` for new parameter candidates to evaluate next."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        rng_ask, rng_restart = jax.random.split(rng)
        restart_state = self.restart(rng_restart, state, params)
        # Simple tree map - jittable if state dimensions are static
        # Otherwise restart wrapper has to overwrite `ask`
        new_state = jax.tree_map(
            lambda x, y: jax.lax.select(state.restart_state.restart_next, x, y),
            restart_state,
            state,
        )
        # Replace/transfer best member and fitness so far
        strategy_state = new_state.strategy_state.replace(
            best_fitness=state.strategy_state.best_fitness,
            best_member=state.strategy_state.best_member,
        )
        x, strategy_state = self.base_strategy.ask(
            rng_ask, strategy_state, params.strategy_params
        )
        return x, new_state.replace(strategy_state=strategy_state)

    @partial(jax.jit, static_argnums=(0,))
    def tell(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: WrapperState,
        params: Optional[WrapperParams] = None,
    ) -> WrapperState:
        """`tell` performance data for strategy state update."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        strategy_state = self.base_strategy.tell(
            x, fitness, state.strategy_state, params.strategy_params
        )
        state = state.replace(strategy_state=strategy_state)
        restart_next = self.stop(fitness, state, params)
        restart_state = state.restart_state.replace(restart_next=restart_next)
        return WrapperState(strategy_state, restart_state)

    @partial(jax.jit, static_argnums=(0,))
    def stop(
        self, fitness: chex.Array, state: WrapperState, params: WrapperParams
    ) -> bool:
        """Check all stopping criteria & return stopping bool indicator."""
        restart_bool = 0
        # Loop over stop criteria functions - stop if one is fullfilled
        for crit in self.stop_criteria:
            restart_bool += crit(fitness, state, params)

        # Check if minimal number of generations has passed!
        return jnp.logical_and(
            restart_bool > 0, min_gen_criterion(fitness, state, params)
        )

    def restart(
        self,
        rng: chex.PRNGKey,
        state: WrapperState,
        params: WrapperParams,
    ) -> WrapperState:
        """Restart state for next generations."""
        # Copy over important parts of state from previous strategy
        new_strategy_state = self.restart_strategy(rng, state, params)
        new_restart_state = state.restart_state.replace(
            restart_counter=state.restart_state.restart_counter + 1,
            restart_next=False,
        )
        return WrapperState(new_strategy_state, new_restart_state)

    def restart_strategy(
        self,
        rng: chex.PRNGKey,
        state: WrapperState,
        params: WrapperParams,
    ) -> WrapperState:
        """Restart strategy specific new state construction."""
        raise NotImplementedError



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from .cma_es import get_cma_elite_weights
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    p_sigma: chex.Array
    M: chex.Array
    mean: chex.Array
    sigma: float
    c_c: chex.Array
    c_d: chex.Array
    weights_truncated: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    mu_eff: float
    c_1: float
    c_mu: float
    c_sigma: float
    d_sigma: float
    chi_n: float
    mu_w: float
    c_m: float = 1.0
    sigma_init: float = 0.065
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class LM_MA_ES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.5,
        memory_size: int = 10,
        sigma_init: float = 1.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Limited Memory MA-ES (Loshchilov et al., 2017)
        Reference: https://arxiv.org/pdf/1705.06693.pdf
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        assert 0 <= elite_ratio <= 1
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.memory_size = memory_size
        self.strategy_name = "LM_MA_ES"

        # Set core kwargs es_params
        self.sigma_init = sigma_init

        # Robustness for int32 - squaring in hyperparameter calculations
        self.max_dims_sq = jnp.minimum(self.num_dims, 40000)

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        _, weights_truncated, mu_eff, c_1, c_mu = get_cma_elite_weights(
            self.popsize, self.elite_popsize, self.num_dims, self.max_dims_sq
        )

        # lrate for cumulation of step-size control and rank-one update
        c_sigma = (mu_eff + 2) / (self.num_dims + mu_eff + 5)
        d_sigma = (
            1
            + 2
            * jnp.maximum(0, jnp.sqrt((mu_eff - 1) / (self.num_dims + 1)) - 1)
            + c_sigma
        )
        chi_n = jnp.sqrt(self.num_dims) * (
            1.0
            - (1.0 / (4.0 * self.num_dims))
            + 1.0 / (21.0 * (self.max_dims_sq ** 2))
        )
        mu_w = 1 / jnp.sum(weights_truncated ** 2)
        params = EvoParams(
            mu_eff=mu_eff,
            c_1=c_1,
            c_mu=c_mu,
            c_sigma=c_sigma,
            d_sigma=d_sigma,
            chi_n=chi_n,
            mu_w=mu_w,
            sigma_init=self.sigma_init,
        )
        return params

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        _, weights_truncated, _, _, _ = get_cma_elite_weights(
            self.popsize, self.elite_popsize, self.num_dims, self.max_dims_sq
        )
        c_d = jnp.array(
            [1 / (1.5 ** i * self.num_dims) for i in range(self.memory_size)]
        )
        c_c = jnp.array(
            [
                self.popsize / (4 ** i * self.num_dims)
                for i in range(self.memory_size)
            ]
        )
        c_c = jnp.minimum(c_c, 1.99)

        # Initialize evolution paths & covariance matrix
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            p_sigma=jnp.zeros(self.num_dims),
            sigma=params.sigma_init,
            mean=initialization,
            M=jnp.zeros((self.num_dims, self.memory_size)),
            weights_truncated=weights_truncated,
            c_d=c_d,
            c_c=c_c,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        x = sample(
            rng,
            state.mean,
            state.sigma,
            state.M,
            self.num_dims,
            self.popsize,
            state.c_d,
            state.gen_counter,
        )
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: chex.ArrayTree,
        params: chex.ArrayTree,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Sort new results, extract elite, store best performer
        concat_p_f = jnp.hstack([jnp.expand_dims(fitness, 1), x])
        sorted_solutions = concat_p_f[concat_p_f[:, 0].argsort()]
        # Update mean, isotropic/anisotropic paths, covariance, stepsize
        mean, z_k = update_mean(
            state.mean,
            state.sigma,
            sorted_solutions,
            params.c_m,
            state.weights_truncated,
        )
        p_sigma, norm_p_sigma = update_p_sigma(
            z_k,
            state.p_sigma,
            params.c_sigma,
            params.mu_eff,
            state.weights_truncated,
        )
        M = update_M_matrix(
            state.M,
            z_k,
            state.c_c,
            params.mu_w,
            state.weights_truncated,
        )
        sigma = update_sigma(
            state.sigma,
            norm_p_sigma,
            params.c_sigma,
            params.d_sigma,
            params.chi_n,
        )
        return state.replace(mean=mean, p_sigma=p_sigma, M=M, sigma=sigma)


def update_mean(
    mean: chex.Array,
    sigma: float,
    sorted_solutions: chex.Array,
    c_m: float,
    weights_truncated: chex.Array,
) -> Tuple[chex.Array, chex.Array]:
    """Update mean of strategy."""
    z_k = sorted_solutions[:, 1:] - mean  # ~ N(0, σ^2 C)
    y_k = z_k / sigma  # ~ N(0, C)
    y_w = jnp.sum(y_k.T * weights_truncated, axis=1)
    mean += c_m * sigma * y_w
    return mean, z_k


def update_p_sigma(
    z_k: chex.Array,
    p_sigma: chex.Array,
    c_sigma: float,
    mu_eff: float,
    weights_truncated: chex.Array,
) -> Tuple[chex.Array, float]:
    """Update evolution path for covariance matrix."""
    z_w = jnp.sum(z_k.T * weights_truncated, axis=1)
    p_sigma_new = (1 - c_sigma) * p_sigma + jnp.sqrt(
        c_sigma * (2 - c_sigma) * mu_eff
    ) * z_w
    norm_p_sigma = jnp.linalg.norm(p_sigma_new)
    return p_sigma_new, norm_p_sigma


def update_M_matrix(
    M: chex.Array,
    z_k: chex.Array,
    c_c: chex.Array,
    mu_w: float,
    weights_truncated: chex.Array,
) -> chex.Array:
    """Update the M matrix."""
    weighted_elite = jnp.sum(
        jnp.array([w * z for w, z in zip(weights_truncated, z_k)]),
        axis=0,
    )
    # Loop over individual memory components - this could be vectorized!
    for i in range(M.shape[1]):
        new_m = (1 - c_c[i]) * M[:, i] + jnp.sqrt(
            mu_w * c_c[i] * (2 - c_c[i])
        ) * weighted_elite
        M = M.at[:, i].set(new_m)
    return M


def update_sigma(
    sigma: float,
    norm_p_sigma: float,
    c_sigma: float,
    d_sigma: float,
    chi_n: float,
) -> float:
    """Update stepsize sigma."""
    sigma_new = sigma * jnp.exp(
        (c_sigma / d_sigma) * (norm_p_sigma / chi_n - 1)
    )
    return sigma_new


def sample(
    rng: chex.PRNGKey,
    mean: chex.Array,
    sigma: float,
    M: chex.Array,
    n_dim: int,
    pop_size: int,
    c_d: chex.Array,
    gen_counter: int,
) -> chex.Array:
    """Jittable Gaussian Sample Helper."""
    z = jax.random.normal(rng, (n_dim, pop_size))  # ~ N(0, I)
    for j in range(M.shape[1]):
        update_bool = gen_counter > j
        new_z = (1 - c_d[j]) * z + (c_d[j] * M[:, j])[:, jnp.newaxis] * (
            M[:, j][:, jnp.newaxis] * z
        )
        z = jax.lax.select(update_bool, new_z, z)
    z = jnp.swapaxes(z, 1, 0)
    x = mean + sigma * z  # ~ N(m, σ^2 C)
    return x



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from ..core import GradientOptimizer, OptState, OptParams, exp_decay


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    opt_state: OptState
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    opt_params: OptParams
    sigma_init: float = 0.03
    sigma_decay: float = 0.999
    sigma_limit: float = 0.01
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class ARS(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.1,
        opt_name: str = "adam",
        lrate_init: float = 0.05,
        lrate_decay: float = 1.0,
        lrate_limit: float = 0.001,
        sigma_init: float = 0.03,
        sigma_decay: float = 1.0,
        sigma_limit: float = 0.01,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Augmented Random Search (Mania et al., 2018)
        Reference: https://arxiv.org/pdf/1803.07055.pdf"""
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        assert not self.popsize & 1, "Population size must be even"
        # ARS performs antithetic sampling & allows you to select
        # "b" elite perturbation directions for the update
        assert 0 <= elite_ratio <= 1
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize / 2 * self.elite_ratio))
        assert opt_name in ["sgd", "adam", "rmsprop", "clipup", "adan"]
        self.optimizer = GradientOptimizer[opt_name](self.num_dims)
        self.strategy_name = "ARS"

        # Set core kwargs es_params (lrate/sigma schedules)
        self.lrate_init = lrate_init
        self.lrate_decay = lrate_decay
        self.lrate_limit = lrate_limit
        self.sigma_init = sigma_init
        self.sigma_decay = sigma_decay
        self.sigma_limit = sigma_limit

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        opt_params = self.optimizer.default_params.replace(
            lrate_init=self.lrate_init,
            lrate_decay=self.lrate_decay,
            lrate_limit=self.lrate_limit,
        )
        return EvoParams(
            opt_params=opt_params,
            sigma_init=self.sigma_init,
            sigma_decay=self.sigma_decay,
            sigma_limit=self.sigma_limit,
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )

        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init,
            opt_state=self.optimizer.initialize(params.opt_params),
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        # Antithetic sampling of noise
        z_plus = jax.random.normal(
            rng,
            (int(self.popsize / 2), self.num_dims),
        )
        z = jnp.concatenate([z_plus, -1.0 * z_plus])
        x = state.mean + state.sigma * z
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Reconstruct noise from last mean/std estimates
        noise = (x - state.mean) / state.sigma
        noise_1 = noise[: int(self.popsize / 2)]
        fit_1 = fitness[: int(self.popsize / 2)]
        fit_2 = fitness[int(self.popsize / 2) :]
        elite_idx = jnp.minimum(fit_1, fit_2).argsort()[: self.elite_popsize]

        fitness_elite = jnp.concatenate([fit_1[elite_idx], fit_2[elite_idx]])
        # Add small constant to ensure non-zero division stability
        sigma_fitness = jnp.std(fitness_elite) + 1e-05
        fit_diff = fit_1[elite_idx] - fit_2[elite_idx]
        fit_diff_noise = jnp.dot(noise_1[elite_idx].T, fit_diff)
        theta_grad = 1.0 / (self.elite_popsize * sigma_fitness) * fit_diff_noise

        # Grad update using optimizer instance - decay lrate if desired
        mean, opt_state = self.optimizer.step(
            state.mean, theta_grad, state.opt_state, params.opt_params
        )
        opt_state = self.optimizer.update(opt_state, params.opt_params)
        sigma = exp_decay(state.sigma, params.sigma_decay, params.sigma_limit)
        return state.replace(mean=mean, sigma=sigma, opt_state=opt_state)



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages


import re
import os
from typing import List

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(CURRENT_DIR, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def parse_requirements(path: str) -> List[str]:
    with open(os.path.join(CURRENT_DIR, path)) as f:
        return [
            line.rstrip() for line in f if not (line.isspace() or line.startswith("#"))
        ]


VERSIONFILE = "evosax/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))
git_tar = f"https://github.com/RobertTLange/evosax/archive/v{verstr}.tar.gz"

requires = (
    [
        "jax",
        "jaxlib>=0.1.37",
        "chex",
        "flax",
        "numpy",
        "pyyaml",
        "pickle5; python_version < '3.8'",
        "matplotlib",
        "dotmap",
    ],
)

setup(
    name="evosax",
    version=verstr,
    author="Robert Tjarko Lange",
    author_email="robertlange0@gmail.com",
    description="JAX-Based Evolution Strategies",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RobertTLange/evosax",
    download_url=git_tar,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    python_requires=">=3.7",
    install_requires=requires,
    tests_requre=["brax", "torch", "torchvision", "gymnax"],
)



import jax
from evosax import Strategies
from evosax.problems import BBOBFitness


def test_strategy_ask(strategy_name):
    # Loop over all strategies and test ask API
    rng = jax.random.PRNGKey(0)
    if strategy_name == "ESMC":
        popsize = 21
    else:
        popsize = 20
    strategy = Strategies[strategy_name](popsize=popsize, num_dims=2)
    params = strategy.default_params
    state = strategy.initialize(rng, params)
    x, state = strategy.ask(rng, state, params)
    assert x.shape[0] == popsize
    assert x.shape[1] == 2
    return


def test_strategy_ask_tell(strategy_name):
    # Loop over all strategies and test ask API
    rng = jax.random.PRNGKey(0)
    if strategy_name == "ESMC":
        popsize = 21
    else:
        popsize = 20
    strategy = Strategies[strategy_name](popsize=popsize, num_dims=2)
    params = strategy.default_params
    state = strategy.initialize(rng, params)
    x, state = strategy.ask(rng, state, params)
    evaluator = BBOBFitness("Sphere", num_dims=2)
    fitness = evaluator.rollout(rng, x)
    state = strategy.tell(x, fitness, state, params)
    return



import chex
import jax
import jax.numpy as jnp
from flax import linen as nn
from ..core.fitness import z_score_trafo, range_norm_trafo


def tanh_age(x: chex.Array, gen_counter: float) -> chex.Array:
    """Return normalized tanh age."""
    return jnp.tanh(x / jnp.float32(gen_counter) - 1.0)


class MultiHeadSelfAttention(nn.Module):
    num_heads: int = 1
    num_features: int = 16

    @nn.compact
    def __call__(self, x: chex.Array) -> chex.Array:
        """Applies multi-head dot product self-attention on the input data.
        Args:
          x: input of shape `[length, features_in]`.

        Returns:
            output of shape `[length, num_features]`.
        """
        assert (
            self.num_features % self.num_heads == 0
        ), "Memory dimension must be divisible by number of heads."
        head_dim = self.num_features // self.num_heads
        # project x to multi-headed q/k/v + apply dot-product attention
        # dimensions are then [length, n_heads, n_features_per_head]
        query = multi_head_embedding(x, self.num_heads, head_dim, "query")
        key = multi_head_embedding(x, self.num_heads, head_dim, "key")
        value = multi_head_embedding(x, self.num_heads, head_dim, "value")
        x_att = scaled_dot_product(query, key, value)
        # Only apply out mixing of heads if more than one head - else squeeze
        if self.num_heads > 1:
            out = mix_head_outputs(x_att, self.num_features, "out")
        else:
            out = jnp.squeeze(x_att, axis=1)
        return out


class MultiHeadCrossAttention(nn.Module):
    num_heads: int = 1
    num_features: int = 16

    @nn.compact
    def __call__(self, x: chex.Array, y: chex.Array) -> chex.Array:
        """Applies multi-head dot product self-attention on the input data.
        Args:
          x: input of shape `[length_1, features_in]`. - Key/value input.
          y: input of shape `[length_2, features_in]`. - Query input.

        Returns:
            output of shape `[length_2, num_features]`.
        """
        assert (
            self.num_features % self.num_heads == 0
        ), "Memory dimension must be divisible by number of heads."
        head_dim = self.num_features // self.num_heads
        # project x to multi-headed q/k/v + apply dot-product attention
        # dimensions are then [length, n_heads, n_features_per_head]
        query = multi_head_embedding(y, self.num_heads, head_dim, "query")
        key = multi_head_embedding(x, self.num_heads, head_dim, "key")
        value = multi_head_embedding(x, self.num_heads, head_dim, "value")
        x_att = scaled_dot_product(query, key, value)
        # Only apply out mixing of heads if more than one head - else squeeze
        if self.num_heads > 1:
            out = mix_head_outputs(x_att, self.num_features, "out")
        else:
            out = jnp.squeeze(x_att, axis=1)
        return out


def multi_head_embedding(
    x: chex.Array, num_heads: int, head_dim: int, label: str
) -> chex.Array:
    """Simple dense general embedding layer."""
    return nn.linear.DenseGeneral(
        features=(num_heads, head_dim),
        use_bias=True,
        name=label,
    )(x)


def mix_head_outputs(
    x: chex.Array, num_features: int, label: str
) -> chex.Array:
    """Simple dense mixing of heads layer."""
    return nn.linear.DenseGeneral(
        features=num_features,
        axis=(-2, -1),
        use_bias=True,
        name=label,
    )(x)


def scaled_dot_product(
    q: chex.Array, k: chex.Array, v: chex.Array
) -> chex.Array:
    """
    Computes dot-product attention given multi-headed query, key, and value.

    Args:
        q - queries for calculating attention with shape of
            `[length, heads, embed_dim]`.
        k - keys for calculating attention with shape of
            `[length, heads, embed_dim]`.
        v - values for calculating attention with shape of
            `[length, heads, embed_dim]`.
    Returns:
        output of shape [length, heads, embed_dim]
    """
    d_k = q.shape[-1]
    attn_logits = jnp.matmul(
        jnp.transpose(q, (1, 0, 2)),
        jnp.transpose(k, (1, 2, 0)),
    )
    attn_logits = attn_logits / jnp.sqrt(d_k)
    attention = nn.softmax(attn_logits, axis=-1)
    values = jnp.matmul(attention, jnp.swapaxes(v, 0, 1))
    return jnp.swapaxes(values, 0, 1)  # [h, l, d] -> [l, h, d]


class SamplingAttention(nn.Module):
    num_att_heads: int
    att_hidden_dims: int

    @nn.compact
    def __call__(self, F_E: chex.Array) -> chex.Array:
        # Perform cross-attention between kids and parents
        S = MultiHeadSelfAttention(self.num_att_heads, self.att_hidden_dims)(
            F_E
        )
        logits = nn.Dense(1)(S)
        return nn.softmax(logits.squeeze(axis=-1))


class SelectionAttention(nn.Module):
    num_att_heads: int
    att_hidden_dims: int

    @nn.compact
    def __call__(
        self, rng: chex.PRNGKey, F_X: chex.Array, F_E: chex.Array
    ) -> chex.Array:
        # Perform cross-attention between kids and parents
        A = MultiHeadCrossAttention(self.num_att_heads, self.att_hidden_dims)(
            F_X, F_E
        )
        # Construct raw selection matrix with row-wise logits
        queries_S = nn.Dense(self.att_hidden_dims)(A)
        keys_S = nn.Dense(self.att_hidden_dims)(F_X)
        # Selection matrix (elite_popsize, popsize)
        S = (queries_S @ keys_S.T) / jnp.sqrt(self.att_hidden_dims)
        # Selection matrix w. parent (elite_popsize, popsize + 1)
        S_p = jnp.concatenate([S, jnp.ones((S.shape[0], 1))], axis=1)
        # Sample kid id to replace or parent id to keep
        idx = jax.random.categorical(rng, S_p, axis=1)
        S_M = jnp.zeros(S_p.shape).at[jnp.arange(S.shape[0]), idx].set(1)
        # Return mask w/o final column corresponding to parent
        return S_M[:, : F_X.shape[0]]


class MutationAttention(nn.Module):
    num_att_heads: int
    att_hidden_dims: int

    @nn.compact
    def __call__(self, sigma: chex.Array, F: chex.Array) -> chex.Array:
        z_feat = z_score_trafo(sigma)
        norm_feat = range_norm_trafo(sigma)
        conc_inputs = jnp.concatenate([F, z_feat, norm_feat], axis=1)
        M = MultiHeadSelfAttention(self.num_att_heads, self.att_hidden_dims)(
            conc_inputs
        )
        log_var = nn.Dense(1)(M)
        multiplier = jnp.exp(0.5 * log_var)
        sigma_out = sigma * multiplier
        return sigma_out



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    rng: chex.PRNGKey
    mean: chex.Array
    archive: chex.Array
    fitness: chex.Array
    sigma: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    sigma_init: float = 0.07
    sigma_meta: float = 2.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class GESMR_GA(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.5,
        sigma_ratio: float = 0.5,
        sigma_init: float = 0.07,
        sigma_meta: float = 2.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Group Elite Selection of Mutation Rates (GESMR) GA."""

        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.num_sigma_groups = int(jnp.sqrt(self.popsize))
        self.members_per_group = int(
            jnp.ceil(self.popsize / self.num_sigma_groups)
        )
        self.sigma_ratio = sigma_ratio
        self.sigma_popsize = max(
            1, int(self.num_sigma_groups * self.sigma_ratio)
        )
        self.strategy_name = "GESMR_GA"
        # Set core kwargs es_params
        self.sigma_init = sigma_init
        self.sigma_meta = sigma_meta

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(sigma_init=self.sigma_init, sigma_meta=self.sigma_meta)

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the differential evolution strategy."""
        rng, rng_init = jax.random.split(rng)
        initialization = jax.random.uniform(
            rng_init,
            (self.elite_popsize, self.num_dims),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            rng=rng,
            mean=initialization[0],
            archive=initialization,
            fitness=jnp.zeros(self.elite_popsize) + jnp.finfo(jnp.float32).max,
            sigma=jnp.zeros(self.num_sigma_groups) + params.sigma_init,
            best_member=initialization[0],
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        rng, rng_idx, rng_eps_x, rng_eps_s = jax.random.split(rng, 4)
        # Sample noise for mutation of x and sigma
        eps_x = jax.random.normal(rng_eps_x, (self.popsize, self.num_dims))
        eps_s = jax.random.uniform(
            rng_eps_s, (self.num_sigma_groups,), minval=-1, maxval=1
        )

        # Sample members to evaluate from parent archive
        idx = jax.random.choice(
            rng_idx, jnp.arange(self.elite_popsize), (self.popsize - 1,)
        )
        x = jnp.concatenate([state.archive[0][None, :], state.archive[idx]])

        # Store fitness before perturbation (used to compute meta-fitness)
        fitness_mem = jnp.concatenate(
            [state.fitness[0][None], state.fitness[idx]]
        )

        # Apply sigma mutation on group level -> repeat for popmember broadcast
        sigma_perturb = state.sigma * params.sigma_meta ** eps_s
        sigma_repeated = jnp.repeat(sigma_perturb, self.members_per_group)[
            : self.popsize
        ]
        sigma = jnp.concatenate([state.sigma[0][None], sigma_repeated[1:]])

        # Apply x mutation -> scale specific to group membership
        x += sigma[:, None] * eps_x
        return x, state.replace(
            archive=x, fitness=fitness_mem, sigma=sigma_perturb
        )

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        # Select best x members
        idx = jnp.argsort(fitness)[: self.elite_popsize]
        archive = x[idx]

        # Select best sigma based on function value improvement
        group_ids = jnp.repeat(
            jnp.arange(self.members_per_group), self.num_sigma_groups
        )[: self.popsize]
        delta_fitness = fitness - state.fitness

        best_deltas = []
        for k in range(self.num_sigma_groups):
            sub_mask = group_ids == k
            sub_delta = (
                sub_mask * delta_fitness
                + (1 - sub_mask) * jnp.finfo(jnp.float32).max
            )
            max_sub_delta = jnp.min(sub_delta)
            best_deltas.append(max_sub_delta)

        idx_select = jnp.argsort(jnp.array(best_deltas))[: self.sigma_popsize]
        sigma_elite = state.sigma[idx_select]

        # Resample sigmas with replacement
        rng, rng_sigma = jax.random.split(state.rng)
        idx_s = jax.random.choice(
            rng_sigma,
            jnp.arange(self.sigma_popsize),
            (self.num_sigma_groups - 1,),
        )
        sigma = jnp.concatenate([state.sigma[0][None], sigma_elite[idx_s]])

        # Set mean to best member seen so far
        improved = fitness[0] < state.best_fitness
        best_mean = jax.lax.select(improved, archive[0], state.best_member)
        return state.replace(
            rng=rng,
            fitness=fitness[idx],
            archive=archive,
            sigma=sigma,
            mean=best_mean,
        )



from .evo_transformer import EvoTransformer
from .features import (
    FitnessFeaturizer,
    FitnessFeaturesState,
    SolutionFeaturizer,
    SolutionFeaturesState,
    DistributionFeaturizer,
    DistributionFeaturesState,
)

__all__ = [
    "EvoTransformer",
    "FitnessFeaturizer",
    "FitnessFeaturesState",
    "SolutionFeaturizer",
    "SolutionFeaturesState",
    "DistributionFeaturizer",
    "DistributionFeaturesState",
]



from typing import Optional
from flax import linen as nn
import jax
import jax.numpy as jnp
import numpy as np
import chex


def scaled_dot_product(
    q: chex.Array,
    k: chex.Array,
    mask: Optional[chex.Array] = None,
) -> chex.Array:
    d_k = q.shape[-1]
    attn_logits = jnp.matmul(q, jnp.swapaxes(k, -2, -1))
    attn_logits = attn_logits / jnp.sqrt(d_k)
    if mask is not None:
        attn_logits = jnp.where(mask == 0, -9e15, attn_logits)
    attention = nn.softmax(attn_logits, axis=-1)
    return attention


def expand_mask(mask: chex.Array) -> chex.Array:
    assert mask.ndim >= 2
    if mask.ndim == 3:
        mask = jnp.expand_dims(mask, axis=1)
    while mask.ndim < 4:
        mask = jnp.expand_dims(mask, axis=0)
    return mask


class MLP(nn.Module):
    embed_dim: int
    dropout_prob: float
    use_bias: bool

    def setup(self):
        self.linear = [
            nn.Dense(
                features=4 * self.embed_dim,
                kernel_init=nn.initializers.xavier_uniform(),
                bias_init=nn.initializers.zeros,
                use_bias=self.use_bias,
            ),
            nn.gelu,
            nn.Dense(
                features=self.embed_dim,
                kernel_init=nn.initializers.xavier_uniform(),
                bias_init=nn.initializers.zeros,
                use_bias=self.use_bias,
            ),
            nn.Dropout(self.dropout_prob),
        ]

    def __call__(self, x: chex.Array, train: bool = True) -> chex.Array:
        for l in self.linear:
            x = l(x) if not isinstance(l, nn.Dropout) else l(x, deterministic=not train)
        return x


class PositionalEncoding(nn.Module):
    d_model: int
    max_len: int = 5000

    def setup(self):
        pe = np.zeros((self.max_len, self.d_model))
        position = np.arange(0, self.max_len, dtype=np.float32)[:, None]
        div_term = np.exp(
            np.arange(0, self.d_model, 2, dtype=np.float32)
            * -(np.log(10000.0) / self.d_model)
        )
        pe[:, 0::2] = np.sin(position * div_term)
        pe[:, 1::2] = np.cos(position * div_term)
        pe = pe[None]
        self.pe = jax.device_put(pe)

    def __call__(self, x: chex.Array) -> chex.Array:
        x = x + self.pe[:, : x.shape[1]]
        return x



import jax
import jax.numpy as jnp
from flax import linen as nn
import chex
from jax.nn.initializers import (
    glorot_normal,
    glorot_uniform,
    he_normal,
    he_uniform,
    kaiming_normal,
    kaiming_uniform,
    lecun_normal,
    lecun_uniform,
    xavier_normal,
    xavier_uniform,
)


kernel_init_fn = {
    "glorot_normal": glorot_normal,
    "glorot_uniform": glorot_uniform,
    "he_normal": he_normal,
    "he_uniform": he_uniform,
    "kaiming_normal": kaiming_normal,
    "kaiming_uniform": kaiming_uniform,
    "lecun_normal": lecun_normal,
    "lecun_uniform": lecun_uniform,
    "xavier_normal": xavier_normal,
    "xavier_uniform": xavier_uniform,
}


def default_bias_init(scale=0.05):
    return nn.initializers.uniform(scale)


def identity_out(
    x: chex.Array, num_output_units: int, init_type: str = "lecun_normal"
) -> chex.Array:
    """Simple affine layer."""
    x_out = nn.Dense(
        features=num_output_units,
        kernel_init=kernel_init_fn[init_type](),
        bias_init=default_bias_init(),
    )(x)
    return x_out


def tanh_out(
    x: chex.Array, num_output_units: int, init_type: str = "lecun_normal"
) -> chex.Array:
    """Simple affine layer & tanh rectification."""
    x = nn.Dense(
        features=num_output_units,
        kernel_init=kernel_init_fn[init_type](),
        bias_init=default_bias_init(),
    )(x)
    return nn.tanh(x)


def categorical_out(
    rng: chex.PRNGKey,
    x: chex.Array,
    num_output_units: int,
    init_type: str = "lecun_normal",
) -> chex.Array:
    """Simple affine layer & categorical sample from logits."""
    x = nn.Dense(
        features=num_output_units,
        kernel_init=kernel_init_fn[init_type](),
        bias_init=default_bias_init(),
    )(x)
    x_out = jax.random.categorical(rng, x)
    return x_out


def gaussian_out(
    rng: chex.PRNGKey,
    x: chex.Array,
    num_output_units: int,
    init_type: str = "lecun_normal",
) -> chex.Array:
    """Simple affine layers for mean and log var + gaussian sample."""
    x_mean = nn.Dense(
        features=num_output_units,
        kernel_init=kernel_init_fn[init_type](),
        bias_init=default_bias_init(),
    )(x)
    x_log_var = nn.Dense(
        features=1,
        kernel_init=kernel_init_fn[init_type](),
        bias_init=default_bias_init(),
    )(x)
    x_std = jnp.exp(0.5 * x_log_var)
    noise = x_std * jax.random.normal(rng, (num_output_units,))
    return x_mean + noise



import chex
import jax
import jax.numpy as jnp


def get_rotation(rng: chex.PRNGKey, num_dims: int) -> chex.Array:
    """Return orthonormal rotation matrix via QR decomposition."""
    # Used to turn separable functions into non-separable
    A = jax.random.normal(rng, (num_dims, num_dims))
    Q = jnp.linalg.qr(A)[0]
    return Q


def lambda_alpha_trafo(alpha: float, num_dims: int) -> chex.Array:
    """LambdaAlpha matrix. Diagonal matrix with alpha powers."""

    def get_diag(i):
        exp = (0.5 * (i / (num_dims - 1))) * (num_dims > 1) + 0.5 * (
            num_dims <= 1
        )
        return alpha ** exp

    diag_vals = jax.vmap(get_diag, in_axes=0)(jnp.arange(num_dims))
    return jnp.diag(diag_vals)


def oscillation_trafo(element: float) -> chex.Array:
    """Oscillation trafo function for x array input & f value output.
    (p.3; Hansen et al., 2009)"""
    x_carat = jax.lax.select(element == 0, 0.0, jnp.log(jnp.abs(element)))
    c1 = jax.lax.select(element > 0, 10.0, 5.5)
    c2 = jax.lax.select(element > 0, 7.9, 3.1)
    return jnp.sign(element) * jnp.exp(
        x_carat + 0.049 * (jnp.sin(c1 * x_carat) + jnp.sin(c2 * x_carat))
    )


def asymmetry_trafo(
    vector: chex.Array, beta: float, num_dims: int
) -> chex.Array:
    """Assymmetry trafo function for x array input & f value output.
    (p.3; Hansen et al., 2009)"""
    dim = vector.shape[0]

    def get_asy_val(idx, val):
        t = jax.lax.select(num_dims > 1, idx / (num_dims - 1.0), 1.0)
        exp_l0 = 1 + beta * t * (val ** 0.5)
        exp = jax.lax.select(val > 0, exp_l0, 1.0)
        return val ** exp

    return jax.vmap(get_asy_val, in_axes=(0, 0))(jnp.arange(dim), vector)


def boundary_penalty(vector: chex.Array, num_dims: int) -> chex.Array:
    """Penalty for large function deviations to ensure boundary handling.
    (p.3; Hansen et al., 2009)"""
    out = jnp.abs(vector) - 5.0
    mask = jnp.arange(out.shape[0]) < num_dims
    return jnp.sum(jnp.maximum(0.0, out * mask) ** 2)



from .fitness import FitnessFeaturizer, FitnessFeaturesState
from .solution import SolutionFeaturizer, SolutionFeaturesState
from .distribution import DistributionFeaturizer, DistributionFeaturesState

__all__ = [
    "FitnessFeaturizer",
    "FitnessFeaturesState",
    "SolutionFeaturizer",
    "SolutionFeaturesState",
    "DistributionFeaturizer",
    "DistributionFeaturesState",
]



from typing import List, Tuple
from functools import partial
import jax.numpy as jnp
from flax import linen as nn
import chex
from .attention import AttentionEncoder
from .perceiver import PerceiverEncoder


class CompressionPerceiver(nn.Module):
    num_latents: int
    latent_dim: int
    embed_dim: int
    num_heads: int
    num_layers: int = 1
    dropout_prob: float = 0.0
    input_dropout_prob: float = 0.0
    use_bias: bool = False
    out_att_maps: bool = False

    def setup(self):
        self.cross_attn_population = partial(
            PerceiverEncoder,
            num_latents=self.num_latents,
            latent_dim=self.latent_dim,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_layers=self.num_layers,
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )
        self.lift_cross = nn.vmap(
            self.cross_attn_population,
            variable_axes={"params": None},
            split_rngs={"params": False, "dropout": True},
            in_axes=(0, None, None, None),
            out_axes=0,
        )

    @nn.compact
    def __call__(
        self, x: chex.Array, train: bool = False
    ) -> Tuple[chex.Array, List[chex.Array]]:
        x = x.transpose(1, 0, 2, 3)
        out, att = self.lift_cross(name="CompressionPerceiver")(
            x,
            None,
            False,
            train,
        )
        out = out.transpose(1, 0, 2, 3)
        if self.out_att_maps:
            att = [jnp.array(a).transpose(1, 0, 2, 3, 4) for a in att]
        return out, att


class SolutionPerceiver(nn.Module):
    num_latents: int
    latent_dim: int
    embed_dim: int
    num_heads: int
    num_layers: int = 1
    dropout_prob: float = 0.0
    input_dropout_prob: float = 0.0
    use_bias: bool = False
    out_att_maps: bool = False

    def setup(self):
        self.cross_attn_population = partial(
            CompressionPerceiver,
            num_latents=self.num_latents,
            latent_dim=self.latent_dim,
            embed_dim=self.embed_dim,
            num_heads=self.num_heads,
            num_layers=self.num_layers,
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )
        self.lift_cross = nn.vmap(
            self.cross_attn_population,
            variable_axes={"params": None},
            split_rngs={"params": False, "dropout": False},
            in_axes=(0, None),
            out_axes=0,
        )

    @nn.compact
    def __call__(
        self, x: chex.Array, train: bool = False
    ) -> Tuple[chex.Array, List[chex.Array]]:
        x = x.transpose(3, 0, 1, 2, 4)
        out, att = self.lift_cross(name="SolutionPerceiver")(
            x,
            train,
        )
        out = out.transpose(1, 2, 3, 0, 4)
        if self.out_att_maps:
            att = [jnp.array(a).transpose(1, 2, 0, 3, 4, 5) for a in att]
        return out, att


class DistributionAttention(nn.Module):
    embed_dim: int
    num_heads: int
    num_layers: int = 1
    dropout_prob: float = 0.0
    input_dropout_prob: float = 0.0
    use_bias: bool = False
    out_att_maps: bool = False

    def setup(self):
        self.transformer = partial(
            AttentionEncoder,
            num_heads=self.num_heads,
            embed_dim=self.embed_dim,
            num_layers=self.num_layers,
            dropout_prob=self.dropout_prob,
            input_dropout_prob=self.input_dropout_prob,
            use_bias=self.use_bias,
            out_att_maps=self.out_att_maps,
        )

        self.lift_att = nn.vmap(
            self.transformer,
            variable_axes={"params": None},
            split_rngs={"params": False, "dropout": False},
            in_axes=(0, None, None, None),
            out_axes=0,
        )

    @nn.compact
    def __call__(
        self, x: chex.Array, train: bool = True
    ) -> Tuple[chex.Array, List[chex.Array]]:
        x = x.transpose(1, 0, 2, 3)
        out, att = self.lift_att(name="DistributionAttention")(x, None, False, train)
        out = out.transpose(1, 0, 2, 3)
        if self.out_att_maps:
            att = jnp.array(att).transpose(2, 1, 0, 3, 4, 5)
        return out, att


class DistributionUpdateNetwork(nn.Module):
    embed_dim: int
    dropout_prob: float = 0.0
    use_bias: bool = False

    def setup(self):
        self.output_net = [
            nn.Dense(
                features=self.embed_dim,
                kernel_init=nn.initializers.xavier_uniform(),
                bias_init=nn.initializers.zeros,
                use_bias=self.use_bias,
            ),
            nn.LayerNorm(self.use_bias),
            nn.relu,
            nn.Dropout(self.dropout_prob),
            nn.Dense(
                features=2,
                kernel_init=nn.initializers.xavier_uniform(),
                bias_init=nn.initializers.zeros,
                use_bias=self.use_bias,
            ),
        ]

    @nn.compact
    def __call__(self, x: chex.Array, train: bool = False) -> chex.Array:
        out = x
        for l in self.output_net:
            out = (
                l(out)
                if not isinstance(l, nn.Dropout)
                else l(out, deterministic=not train)
            )
        return out



import jax.numpy as jnp
from evosax import FitnessShaper


def test_fitness_shaping_rank():
    shaper = FitnessShaper(centered_rank=True)
    x = jnp.array([[1.0], [2.0], [3.0]])
    fit = jnp.array([0.0, 1.0, 2.0])
    fit_re = shaper.apply(x, fit)
    assert jnp.allclose(fit_re, jnp.array([-0.5, 0.0, 0.5]), atol=1e-03)


def test_fitness_shaping_decay():
    shaper = FitnessShaper(w_decay=0.01)
    x = jnp.array([[1.0], [2.0], [3.0]])
    fit = jnp.array([0.0, 1.0, 2.0])
    fit_re = shaper.apply(x, fit)
    assert jnp.allclose(
        fit_re, jnp.array([0.01, 1 + 0.04, 2 + 0.09]), atol=1e-03
    )


def test_fitness_shaping_zscore():
    shaper = FitnessShaper(z_score=True)
    x = jnp.array([[1.0], [2.0], [3.0]])
    fit = jnp.array([0.0, 1.0, 2.0])
    fit_re = shaper.apply(x, fit)
    assert jnp.allclose(
        fit_re, jnp.array([-1.2247448, 0.0, 1.2247448]), atol=1e-03
    )


def test_fitness_shaping_max():
    shaper = FitnessShaper(w_decay=0.01, maximize=True)
    x = jnp.array([[1.0], [2.0], [3.0]])
    fit = jnp.array([0.0, 1.0, 2.0])
    fit_re = shaper.apply(x, fit)
    assert jnp.allclose(
        fit_re, jnp.array([0.01, -(1 - 0.04), -(2 - 0.09)]), atol=1e-03
    )
